/*
  Name: Tiny pong clone
  Copyright: GPL
  Author: Andr Schnabel
  Date: 21.02.06 18:05
*/

// pong.c

#include <SDL/SDL.h>
#include <SDL/SDL_mixer.h>
#include <string.h>
#include <math.h>

//==============================================================================
//
// DEFINES
//
//==============================================================================

#define CAPTION			"Pong - stupid clone by Andre Schnabel"

#define SCR_W			640
#define SCR_H			480
#define SCR_BPP			16

#define PADDLE_SPEED	8
#define BALL_SPEED		12

// Math stuff
#define PI				M_PI
#define ARC2RAD(arc) 	(arc * PI) / 180
#define SIN(alpha)		sin(ARC2RAD(alpha))
#define COS(alpha)		cos(ARC2RAD(alpha))

#define RESET_PLAYERS()	players[0].paddle.y = (SCR_H - paddleImg->h) / 2; \
						players[1].paddle.y = players[0].paddle.y;

#define RESET_BALL() 	ball.pos.x = (SCR_W - ballImg->w) / 2; \
						ball.pos.y = (SCR_H - ballImg->h) / 2; \
						flip = !flip; \
						if(flip) \
							ball.movAngle = 0.0f; \
						else \
							ball.movAngle = 180.0f;

#define FADE_SPEED		5

//==============================================================================
//
// TYPES
//
//==============================================================================

// Own boolean type, since we're not using C++
typedef enum { false, true } boolean;

typedef enum
{
	DIR_UP,
	DIR_DOWN
} dir_t;

typedef struct pos_s
{
	int x, y;
} pos_t;

enum
{
	PLAYER1,
	PLAYER2,
	NUM_PLAYERS
};

typedef struct player_s
{
	int score;
	pos_t paddle;
	boolean moving;
	dir_t lMovDir;
} player_t;

typedef struct ball_s
{
	pos_t pos;
	float movAngle;
} ball_t;

//==============================================================================
//
// PROTOTYPES
//
//==============================================================================

SDL_Surface *LoadImage(const char *filename, boolean useAlpha);
void Error(const char *desc);

void Init(void);
void Quit(void);
void ShowSplash(void);
void MainLoop(void);

void Input(void);
void Update(void);
void Draw(void);

//==============================================================================
//
// GLOBAL VARS
//
//==============================================================================

SDL_Surface *screen;
Uint8 *keys;
boolean done;

Uint32 bgColor, black;
SDL_Surface *paddleImg, *ballImg;
SDL_Surface *digitsImg;

Mix_Chunk *ballOut, *ballHits;

player_t players[NUM_PLAYERS];
ball_t ball;
boolean flip = false;

boolean fullscreen = false, bot = false;

//==============================================================================
//
// IMPLEMENTATION
//
//==============================================================================

//
// ParseArgs
//
void ParseArgs(int argc, char **argv)
{
	int i;
	for(i=1; i<argc; i++)
	{
		if(!strcmp(argv[i], "-fullscreen"))
			fullscreen = true;
		
		if(!strcmp(argv[i], "-bot"))
			bot = true;
	}
}

//
// main
//
int main(int argc, char **argv)
{
	ParseArgs(argc, argv);
	Init();
	ShowSplash();
	MainLoop();	
	Quit();
	
	return 0;
}

//
// LoadImage
//
SDL_Surface *LoadImage(const char *filename, boolean useAlpha)
{
	SDL_Surface *temp, *image;
	Uint32 magenta = SDL_MapRGB(screen->format, 255, 0, 255);
	
	temp = SDL_LoadBMP(filename);
	if(!temp)
	{
		printf("Error: Can't load %s!\n", filename);
		fflush(stdout);
		exit(1);
	}
	
	image = SDL_DisplayFormat(temp);
	SDL_FreeSurface(temp);
	
	if(useAlpha)
		SDL_SetColorKey(image, SDL_SRCCOLORKEY, magenta);
	
	return image;
}

//
// Error
//
void Error(const char *desc)
{
	printf("Error: %s\n", desc);
	fflush(stdout);
	exit(1);
}

//
// Init
//
void Init(void)
{
	int result;
	Uint32 scrFlags = SDL_HWSURFACE | SDL_DOUBLEBUF;
	
	result = SDL_Init(SDL_INIT_VIDEO);
	if(result == -1)
		Error("Can't init SDL!");
		
	if(fullscreen)
		scrFlags += SDL_FULLSCREEN;
	
	screen = SDL_SetVideoMode(SCR_W, SCR_H, SCR_BPP, scrFlags);	
	if(!screen)
		Error("Can't set video mode!");
	
	keys = SDL_GetKeyState(NULL);
	
	SDL_WM_SetCaption(CAPTION, NULL);
	
	result = Mix_OpenAudio(44100, AUDIO_S16SYS, 2, 2048);
	if(result == -1)
		Error("Can't open audio stuff!");
	
	bgColor = SDL_MapRGB(screen->format, 100, 0, 0);
	black = SDL_MapRGB(screen->format, 0, 0, 0);
	
	paddleImg = LoadImage("paddle.bmp", true);
	ballImg = LoadImage("ball.bmp", true);
	
	players[0].paddle.x = 0;
	players[0].paddle.y = (SCR_H - paddleImg->h) / 2;
	players[1].paddle.x = SCR_W - paddleImg->w;
	players[1].paddle.y = players[0].paddle.y;
	players[0].moving = players[1].moving = false;
	players[0].lMovDir = players[1].lMovDir = DIR_UP;
	players[0].score = players[1].score = 0;
	
	RESET_BALL();
	RESET_PLAYERS();
	
	digitsImg = LoadImage("digits.bmp", true);
	
	ballOut = Mix_LoadWAV("ballOut.wav");
	ballHits = Mix_LoadWAV("ballHits.wav");
}

//
// Quit
//
void Quit(void)
{
	Mix_FreeChunk(ballHits);
	Mix_FreeChunk(ballOut);
	
	SDL_FreeSurface(digitsImg);
	SDL_FreeSurface(ballImg);
	SDL_FreeSurface(paddleImg);
	
	Mix_CloseAudio();
	SDL_Quit();
}

//
// ShowSplash
//
void ShowSplash(void)
{
	boolean splDone = false;
	SDL_Event event;
	
	SDL_Surface *splashImg = LoadImage("splash.bmp", false);	
	SDL_BlitSurface(splashImg, NULL, screen, NULL);
	SDL_Flip(screen);
	
	while(!splDone)
	{
		while(SDL_PollEvent(&event))
			if(event.type == SDL_QUIT)
				splDone = true;
		
		SDL_PumpEvents();
		
		if(keys[SDLK_ESCAPE] || keys[SDLK_RETURN])
			splDone = true;
	}
			
	SDL_FreeSurface(splashImg);
}

//
// MainLoop
//
void MainLoop(void)
{
	done = false;
	Uint32 lTime;
	
	do
	{
		lTime = SDL_GetTicks();
		
		Input();
		Update();
		Draw();
		
		SDL_Flip(screen);
		
		while(SDL_GetTicks() - lTime < 1000 / 25);
		
	} while(!done);
}

//
// MovePaddle
//
void MovePaddle(int playerNr, dir_t dir)
{
	int movAdd = PADDLE_SPEED;
	
	if(dir == DIR_UP)
		movAdd *= -1;
	
	players[playerNr].paddle.y += movAdd;
	
	if(players[playerNr].paddle.y < 0)
		players[playerNr].paddle.y = 0;
	if(players[playerNr].paddle.y + paddleImg->h > SCR_H)
		players[playerNr].paddle.y = SCR_H - paddleImg->h;
	
	players[playerNr].moving = true;
	players[playerNr].lMovDir = dir;
}

//
// Input
//
void Input(void)
{
	static SDL_Event event;
	
	// General input

	// Handle events	
	while(SDL_PollEvent(&event))
	{
		if(event.type == SDL_QUIT)
			done = true;
	}
	
	SDL_PumpEvents();
	
	if(keys[SDLK_ESCAPE])
		done = true;
		
	// Player 1
	if(keys[SDLK_UP])
		MovePaddle(PLAYER1, DIR_UP);
	
	if(keys[SDLK_DOWN])
		MovePaddle(PLAYER1, DIR_DOWN);
	
	if(!keys[SDLK_UP] && !keys[SDLK_DOWN])
		players[0].moving = false;
	
	// Player 2
	if(keys[SDLK_w])
		MovePaddle(PLAYER2, DIR_UP);
	if(keys[SDLK_s])
		MovePaddle(PLAYER2, DIR_DOWN);
	
	if(!keys[SDLK_w] && !keys[SDLK_s])
		players[1].moving = false;
}

//
// Coll
//
boolean Coll(SDL_Rect *rect1, SDL_Rect *rect2)
{
	return (	rect1->x + rect1->w > rect2->x &&
				rect1->x < rect2->x + rect2->w &&
				rect1->y + rect1->h > rect2->y &&
				rect1->y < rect2->y + rect2->h);
}

//
// Update
//
void Update(void)
{
	// For ball movement
	static int xAdd, yAdd;
	
	// Rectangles for collision detection
	static SDL_Rect ballRect;
	static SDL_Rect p1rect, p2rect;
	
	static int dist;
	
	if(ball.movAngle > 360)
		ball.movAngle -= 360;
	
	if(ball.movAngle < 0)
		ball.movAngle = 360 + ball.movAngle;
	
	xAdd = COS(ball.movAngle) * BALL_SPEED;
	yAdd = SIN(ball.movAngle) * BALL_SPEED;
	
	ball.pos.x += xAdd;
	ball.pos.y += yAdd;
	
	ballRect.x = ball.pos.x;
	ballRect.y = ball.pos.y;
	
	ballRect.w = ballImg->w;
	ballRect.h = ballImg->h;
	
	p1rect.x = players[0].paddle.x;
	p1rect.y = players[0].paddle.y;
	
	p1rect.w = p2rect.w = paddleImg->w;
	p1rect.h = p2rect.h = paddleImg->h;
	
	p2rect.x = players[1].paddle.x;
	p2rect.y = players[1].paddle.y;
	
	// Collision with player 1
	if(Coll(&ballRect, &p1rect))
	{
		ball.pos.x += 10;
		
		if(abs(ball.movAngle - 180) > 5)
		{
			if(ball.movAngle > 90 && ball.movAngle < 180)
				ball.movAngle = 180 - ball.movAngle;
			else if(ball.movAngle > 180 && ball.movAngle < 270)
				ball.movAngle = 360 - (ball.movAngle - 180); /*(270 - ball.movAngle) + 90;*/
		}
		else
		{
			dist = abs((ball.pos.y + ballImg->h / 2) - (p1rect.y + paddleImg->h / 2));
			ball.movAngle = 0;
			
			if(ball.pos.y + ballImg->h / 2 > p1rect.y + paddleImg->h / 2)
				ball.movAngle += dist * 2;
			else
				ball.movAngle -= dist * 2;
		}
		
		Mix_PlayChannel(-1, ballHits, 0);
	}
	// Collision with player 2
	else if(Coll(&ballRect, &p2rect))
	{
		ball.pos.x -= 10;
		
		if(abs(ball.movAngle/* - 0*/) > 5)
		{
			if(ball.movAngle > 0 && ball.movAngle < 90)
				ball.movAngle = 90 - ball.movAngle + 90;
			else if(ball.movAngle > 270 && ball.movAngle < 360)
				ball.movAngle = 360 - ball.movAngle + 180;
		}
		else
		{
			dist = abs((ball.pos.y + ballImg->h / 2) - (p2rect.y + paddleImg->h / 2));
			ball.movAngle = 180;
			
			if(ball.pos.y + ballImg->h / 2 > p2rect.y + paddleImg->h / 2)
				ball.movAngle -= dist * 2;
			else
				ball.movAngle += dist * 2;
		}
		
		Mix_PlayChannel(-1, ballHits, 0);
	}
	
	// Going out left
	if(ballRect.x < 0)
	{
		players[0].score++;
		RESET_PLAYERS();
		RESET_BALL();
		Mix_PlayChannel(-1, ballOut, 0);
		
	}
	// Going out right
	else if(ballRect.x + ballImg->w > SCR_W)
	{
		players[1].score++;
		RESET_PLAYERS();
		RESET_BALL();
		Mix_PlayChannel(-1, ballOut, 0);
	}
	
	// Hitting the higher border
	if(ballRect.y < 10)
	{
		ball.pos.y = 10;
		
		if(ball.movAngle > 180 && ball.movAngle < 270)
//			ball.movAngle = 180 - (270 - ball.movAngle);
			ball.movAngle = 180 - (ball.movAngle - 180);
		else
			ball.movAngle = 90 - (ball.movAngle - 270);
		
		Mix_PlayChannel(-1, ballHits, 0);
	}
	
	// Hitting the lower border
	if(ballRect.y + ballImg->h > SCR_H - 10)
	{
		ball.pos.y = SCR_H - 10 - ballImg->h;
		
		if(ball.movAngle > 90 && ball.movAngle < 180)
			ball.movAngle = 270 - (ball.movAngle - 90);
		else
			ball.movAngle = 360 - ball.movAngle;
		
		Mix_PlayChannel(-1, ballHits, 0);
	}
	
	// Bot thinking
	if(bot)
	{
		if(ball.pos.y - 20 < players[1].paddle.y)
			MovePaddle(PLAYER2, DIR_UP);
		else if(ball.pos.y + 20 > players[1].paddle.y + paddleImg->h)
			MovePaddle(PLAYER2, DIR_DOWN);		
	}
}

//
// DrawImage
//
void DrawImage(SDL_Surface *img, int x, int y)
{
	static SDL_Rect rect;
	
	rect.x = x;
	rect.y = y;
	rect.w = img->w;
	rect.h = img->h;
	
	SDL_BlitSurface(img, NULL, screen, &rect);
}

//
// DrawMiddleLine
//
void DrawMiddleLine(void)
{
	static SDL_Rect rect = { SCR_W / 2 - 5, 0, 10, SCR_H };
	SDL_FillRect(screen, &rect, black);	
}

//
// DrawNumber
//
void DrawNumber(int num, int x, int y)
{
	static char str[256];
	static SDL_Rect srcRect, destRect;
	static int length, i;
	
	srcRect.w = destRect.w = 20;
	srcRect.h = destRect.h = 25;
	srcRect.y = 0;
	destRect.x = x;
	destRect.y = y;
	
	sprintf(str, "%d", num);
	length = strlen(str);
	
	for(i=0; i<length; i++)
	{
		destRect.x += destRect.w;
		srcRect.x = (str[i] - '0') * srcRect.w;
		
		SDL_BlitSurface(digitsImg, &srcRect, screen, &destRect);
	}
}

//
// Draw
//
void Draw(void)
{
	int i;
	static int r = 255, g = 255, b = 255;
	static int change = 0;
	
	// Fill background with color
//	SDL_FillRect(screen, NULL, bgColor);	
	SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format, r, g, b));
	
	if(change == 0)
	{
		if(r+FADE_SPEED < 255)
			r+=FADE_SPEED;
		else
			change = 1;
		
		if(g-FADE_SPEED > 0)
			g-=FADE_SPEED;
		if(b-FADE_SPEED > 0)
			b-=FADE_SPEED;
	}
	
	if(change == 1)
	{
		if(g+FADE_SPEED < 255)
			g+=FADE_SPEED;
		else
			change = 2;
		
		if(r-FADE_SPEED > 0)
			r-=FADE_SPEED;
		if(b-10 > 0)
			b-=FADE_SPEED;
	}
	
	if(change == 2)
	{
		if(b+FADE_SPEED < 255)
			b+=FADE_SPEED;
		else
			change = 0;
		
		if(r-FADE_SPEED > 0)
			r-=FADE_SPEED;
		if(g-FADE_SPEED > 0)
			g-=FADE_SPEED;
	}
	
	DrawMiddleLine();
	
	for(i=0; i<NUM_PLAYERS; i++)
		DrawImage(paddleImg, players[i].paddle.x, players[i].paddle.y);
	
	DrawImage(ballImg, ball.pos.x, ball.pos.y);
	
	DrawNumber(players[0].score, SCR_W - 120, 10);
	DrawNumber(players[1].score, 120, 10);
}
